// SDIclass.h

// adi 05/25/99

// this class will communications on the SDI bus
// when incorperated with Tom M.'s code, there will be one instance of this class running
// it will manage the SDI bus as a shared resource
// it will provide the interface functions (described below) to Tom M.s higher level code


/////////////////////////////////////////////////////////////////////////////////
//Interface functions to SDI DLL
//
//This function sends a command string to the SDI bus and returns
//after the initial response from the sensor.  It is intended for non-
//measurement commands and interactive user input.
//
//SDIerror SendCmd(const CString& Cmd, TCHAR * szReturn, DWORD dwBufferSize, DWORD * pdwReturn);
//
//PARAMS: szCmd - fully formatted command string to send out the bus.
//                              This string includes the SDI address etc. (ex. "aI!")
//
//         szReturn - buffer to hold return string.
//
//         dwBufferSize - sizeof the return buffer in bytes.
//
//         pdwReturn - number of valid elements in the return buffer.
//                                              This is needed because there may be successfull
//                                              commands that return no data?? or data with NULLs in it??
//
//RETURN: Error code or success - enum SDIerror
/////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
//
//This function sends a cmd string to the SDI bus and waits for a data response.
//This function does not return until the data is received or a timeout happens.
//
//SDIerror CollectData(const CString& Cmd, TCHAR * pdReturn, DWORD dwBufferSize,
//                                      DWORD * pdwReturn, DWORD dwTimeout);
//
//PARAMS: szCmd - fully formatted command string to send out the bus.
//                              This string includes the address (ex. "aM!")
//
//      pdReturn - pointer to an array of doubles to hold return data.
//
//      dwBufferSize - max size in bytes (not elements) of return buffer.
//
//      pdwReturn - number of valid doubles in return buffer.
//
//      dwTimeout - maximum amount of time in millseconds this function will
//                              wait before returning with an error code.
//
//RETURN:       Error code or success - enum SDIerror
/////////////////////////////////////////////////////////////////////////////////


/////////////////////////////////////////////////////////////////////////////////
//
//  this function aborts a reading called by CollectData (which may be waiting
//  for the SDI bus to become available) - user abort
//
//
//  void Abort() {SetEvent(m_Abort);};
//
//RETURN:   this function does not return a value, but CollectData should return
//          an ABORTED SDIerror
/////////////////////////////////////////////////////////////////////////////////
#ifndef _SDICLASS_H
#define _SDICLASS_H

#include "SerialCommClass.h"        // this is where the serial io is defined
#include "Utils/Semaphore.h"
#include "Utils/Time.h"
#include "Xpert/Win32Fix.h"

#undef DECL
#ifdef SDI_DLL
#define DECL _declspec(dllexport)
#else
#define DECL _declspec(dllimport)
#endif

class DECL CTimer
{
   // this class is used for timeouts
   // eg.  CTimer Timer(20);           // sets up timer to expire 20 ms from now
   //      while(!Timer.Expired);      // loop until timer expires
   // all functions are inline (except for Dgb() which should be removed for the final version

public:
   CTimer(DWORD pDuration) { Ticks=GetTickCount(); Duration=pDuration; };  // pass the timout value in ms
   CTimer(DWORD pDuration, bool expired) {Duration=pDuration; Ticks = GetTickCount() - Duration; };// use this second constructor to start an expired timer
   CTimer& operator=(const CTimer& t) { Duration=t.Duration; Ticks=t.Ticks; return *this; };
   bool Expired() { return (GetTickCount() - Ticks) >= Duration; };    // returns true if timer has expired
   void Expire() { Ticks = GetTickCount() - Duration; };               // use this to force a timer to expire
   void Restart() { Ticks = GetTickCount(); };                         // resets a timer
#ifdef ADI_DEBUG
   void Dbg();                                                         // debug output
#endif
private:
   DWORD Duration;  // Duration specifies how long b4 timer expires,
   DWORD Ticks;     // Ticks holds the time (tickcount) when the timer was started
};


/////////////////////////////////////////////////////////////////////////////
//
// SDI-12 timeout values.

// Note that in SDI-12, the length in time of a single char is 8.33ms and
// the maximum allowable time between characters is 1.66ms (end of stop bit, 
// before next start bit). The SDI spec is inherently more forgiving to 
// the logger. These timings are set to take advantage of that, which makes 
// up for our inability to precisely time serial events.

// A reasonable time to sleep while waiting for new characters, or about 
// half a character time.
const DWORD gc_dwWait = 5;

// The time in ms to wait after receiving a char before declaring a message on 
// the bus has ended. Normally this is simply the inter-char timeout. But, 
// we can't see start bits, so min is 8.33 + 1.66 = 10ms. However, the value 
// must be large enough to account for latency in the driver's processing 
// of chars, which can be high when the user is interacting with the device. 
const DWORD gc_dwLineDead = 35;

// The amount of time in ms that must pass since the last communication with a 
// sensor before we need to send a break.
const DWORD gc_dwLastTx = 87;

// The time in ms that must pass during the inner retry loop to be sure that 
// at least one retry attempt occurs after the 100ms time allowed for a sensor 
// to wake-up. 
const DWORD gc_dwRetry = 125;

// The time to wait in ms for a response from a sensor. The SDI-12 standard 
// specifies 15ms min from stop bit to start bit. Since we can't detect start 
// bits, the minimum becomes 15 plus one char-time of 8.33, or 24ms.
const DWORD gc_dwTimeout = 30;

// The duration of a break in ms (12ms min). 
const DWORD gc_dwBreakTime = 15;

// The duration of a mark in ms (8.33ms min).
const DWORD gc_dwMarkTime = 10;

// The size to allocate an SDI-12 buffer 
const int gc_iMaxSDICmdRespLen = 100;

// The maximum number of characters allowed in any SDI-12 command or response. 
const int gc_iSDICmdRespLen = 80;

// The maximum number of measurements we accept from a sensor.
const int gc_iMaxSDIMeasurements = 40;


/////////////////////////////////////////////////////////////////////////////
//
// SDI-12 error values. 
// Note: SDIerror enum and strError string array must be kept consistent.

enum SDIerror
{
   SDI_OK, BAD_DATA_FROM_SENSOR, BAD_COMMAND_DATA, BUFFER_TOO_SMALL, 
      WAIT_TOO_LONG, ABORTED,CANNOT_OPEN_SDI_PORT, COMM_ERROR, 
      NO_RESPONSE_FROM_SENSOR, DATA_FROM_WRONG_SENSOR, 
      SENSOR_ABORTED_MEASUREMENT, UNKNOWN_ERROR, NUM_SDI_ERRORS
};

extern "C" DECL TCHAR* szSDIError[NUM_SDI_ERRORS];
//  0   SDI_OK - sucess
//  1   BAD_DATA_FROM_SENSOR
//  2   BAD_COMMAND_DATA - error in command string.
//  3   BUFFER_TOO_SMALL - buffer size passed in too small to hold return data.
//  4   WAIT_TOO_LONG - sensor did not respond within calling function's timeout. 
//  5   ABORTED - user abort.
//  6   CANNOT_OPEN_SDI_PORT - the sdi port cannot be opened.
//  7   COMM_ERROR - problems talking to the comm port (eg. parity error).
//  8   NO_RESPONSE_FROM_SENSOR - sensor did not respond even after retries.
//  9   DATA_FROM_WRONG_SENSOR - wrong sensor responded to the request.
//  10  SENSOR_ABORTED_MEASUREMENT - sensor does not have the data.
//  11  UNKNOWN_ERROR - just hope it doesnt happen.


// this enumeration used to denote command type
enum command_type
{
   ACK_ACTIVE, SEND_ID, ADDR_CHANGE, ADDR_QUERY, START_MEASURE, SEND_DATA,
      START_VERIFY, START_CONCURRENT, CONITINOUS_MEASURE, BAD_COMMAND
};

#define _max__highlevel_retries 3
// number of retries not specified by sdi protocol


/////////////////////////////////////////////////////////////////////////////
//
// CSDIComm

class DECL CSDIcomm
{
public:
   // One SDI port - com5 is default.
   CSDIcomm(LPCTSTR CommPort = _T("COM5:"));
   ~CSDIcomm();

   // FUNCTIONS TO BE EXPORTED TO HIGHER LEVEL CODE
public:
   SDIerror SendCmd(const CString& Cmd, TCHAR * szReturn, DWORD dwBufferSize, DWORD * pdwReturn);
   SDIerror CollectData(const CString& Cmd, double * pdReturn, DWORD dwBufferSize, DWORD * pdwReturn, DWORD dwTimeout, BOOL NoWait);  //NoWait is for sensors that support concurrency with the M command.
public:
   void Abort() {SetEvent(m_Abort);};
   void ClearAbort() {ResetEvent(m_Abort);};
   bool ReadSDIBuffer(UCHAR* mbsBuf, int iBufSize);
   CString GetPortName() { return m_SDIport.GetPortName(); };
   CSDIComm m_SDIport;
   static CPtrArray m_SDIPorts;

private:
   bool Initialize();                       // call this to open comm port with sdi settings (and disable fifos)
   bool IsInitialized() {return (m_SDIport.IsOpen());};

   bool str2float(unsigned char * str, int& index_offset, double * float_value); // this function parses the return strings data into doubles

   bool SendCommand(const CString& command, bool SendBreak, DWORD& StartTime);  // sends a command out, including marking
   bool GetResponse(unsigned char * response, bool CheckCRC);  // gets a reply, including ignoring the echoed command in our rx buffer
   bool Raw(const CString& command, unsigned char * response, bool CheckCRC); // sends a command a receives a response to and from sensor

   command_type ParseCommand(const CString& command); // parses the command string to find out what command is to be sent
   bool ParseMeasurementResponse(unsigned char address, unsigned char * response, DWORD& time_till_data, int& measurements_to_take);
   bool GetData(const CString& command, unsigned char * response, double * data_response, int max_measurements, int&  measurements_collected, DWORD ms_timeout, BOOL NoWait, bool CheckCRC);
   bool GetDataFromSensor(unsigned char address, unsigned char * response, double * data_response, int max_measurements, int&  measurements_collected, int measurements_to_take, bool CheckCRC);
   bool GetDataFromSensorLow(unsigned char address, unsigned char * response, double * data_response, int max_measurements, int&  measurements_collected, int measurements_to_take, bool CheckCRC);
   bool WaitOnSensor(DWORD time_till_data, unsigned char address, bool CheckServiceRequest);
   bool RequestDataFromSensor(const CString& command, unsigned char * response, command_type current_command, DWORD& time_till_data, DWORD ms_timeout, int& measurements_to_take);
   bool RequestDataFromSensorLow(const CString& command, unsigned char * response, command_type current_command, DWORD& time_till_data, DWORD ms_timeout, int& measurements_to_take);
   bool IsServiceRequest(unsigned char address);
   bool IsEndOfResponse(unsigned char * ch);

   // Bus lock handling
private:
   HANDLE m_BusFree;
   HANDLE m_Abort;     // UNRESOLVED - abort is to be passed down from high level code
private:
   bool LockBus(UCHAR Addr); // waits for bus to be free, then locks it; returns false if aborted or timedout, true if bus free
   void FreeBus(UCHAR Addr); // unlocks bus

   // Lock used to synchronize Initialize().
   TSemaphore m_semInitLock;

   // DATA MEMBERS
private:
   bool AddressInuse[256]; // Flags to make sure the same address isn't used at the same time
   CTimer LastTxTimer;     // timer - starts expired - we need this one to be global to the class
   int m_poll_sleep;       // how long to sleep (ms) while polling the bus for a service request
   SDIerror m_Error;       // holds last error; error reporting will work like this:
   // functions in this class will return bool to denote sucess
   // if the functions fail, we will return the m_Error, otherwise we will return SDI_OK
   // it is the responsibility of the function that returns false to set m_Error
   // clearing m_Error should not be an issue in this setup as long as m_Error gets set at every place a function can return false
   int m_response_size;    // keeps track of the number of chars received

   /******************
   NOTE
   the way the nextgen sdi is set up in the hardware results in the following:

   before sending data out on the line, RTS must be enabled.
   before receiving data from the line, RTS must be disabled.
   sending data involves raising RTS, sending data, making sure data is out of uart, dropping RTS
   we make sure the data is out of uart by directly checking the uarts status register
   high-level checking is too slow to work

   there is a loopback that results in an echo - ie all data sent out comes back (as if received)
   this data can be used for error checking, but this class treats it like garbage
   so we must make sure to get all the echoed data out of the rx buffer so it does not mix in
   with the data received from the sensors

   the uarts fifos must be disabled - otherwise we wont know if we got a response from the sensor in time
   as the uart will not do an rx interrupt until it receives a set number of bytes (which cannot be lower
   that 8 with the fifos enabled)
   as we are doing 1200 baud comm, the lack of fifos should not be noticed

   Comments from Dan:
   currently the communications class code outputs a character at a time using writefile. I
   tried changing that to do a whole buffer at a time but only the last character appeared.
   Apparently this is because fifo is turned off using:
      _outp(0x7142, (_inp(0x7142) & ~0x01));  // fifo off - therwise the uart waits for a full fifo b4 interrupting on an rx
   When I commented this out, the system would hang on the setcommbreak function.
   There is probably a problem in the driver that prevents it from sending a buffer when the fifo is disabled.
   there is also probably a problem in the driver that makes the setcommbreak hang.
   When I commented out the setcommbreak the code (fifo's enabled) the code did output a string.
   I replaced the test for txbufferempty with the class function waittxend.  that hangs too.
   

   10/16/2001 JDW Switched out Dan's direct H/W I/O and now uses the fifo's and new routine
                  to determine when the last char has been sent.
   
   05/25/2005 JDW Attempted to improve the timing by disabling FIFO's and adding support for this
                  in the kernel driver, but there are still fundamental timing problems with the
                  WaitForSingleObject and Sleep APIs that result in timeouts, so I've gone with some
                  simpler improvements. The allocation size of the SDI-12 buffer has been increased
                  to leave room in case of a buffer overrun, the code now checks for a timeout
                  condition *after* checking for keypresses, rather then before, which will prevent
                  an timeout when a response is actually available, and the code will now check for
                  a service request response after the wait has timed out. These changes are meant to
                  compensate somewhat for the lack of reliable Sleep timing (ie a Sleep(50) actually
                  varied from 60 to 90ms). The thread now runs in "Fast" mode rather then "Standby"
                  to hopefully improve timing. If the system enters "Standby" mode, it will issue a
                  halt which will cause timing to become synced with the 25ms system tick.

   *****************/
};

#endif
